/**
 * EM for GMM actually.
 */
#include "kdtree.hpp"
#include <iostream>
#include <fstream>
#include <vector>
#include <math.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>
#include "vecmath.h"
#define DIM 2

struct kdtreeNode
{
 typedef double value_type;
 int id;
 kdtreeNode():id(0){
    xy[0] = 0;
    xy[1] = 0;
 }
 kdtreeNode(const kdtreeNode& a){
  for(int ii = 0;ii<DIM;ii++){
    xy[ii] = a.xy[ii];
  }
  id = a.id;
 }

  kdtreeNode & operator=(const kdtreeNode& a){
  for(int ii = 0;ii<DIM;ii++){
    xy[ii] = a.xy[ii];
  }
  id = a.id;

    return *this;
  }
 
 double xy[DIM];
 size_t index;

 value_type operator[](size_t n) const
 {
   return xy[n];
 }

 value_type & operator[](size_t n)
 {
   return xy[n];
 }
 
 double distance( const kdtreeNode &node)const
 {
   double x = xy[0] - node.xy[0];
   double y = xy[1] - node.xy[1];
   //only supports manhattan distance
   //whatever
   return std::max(fabs(x),fabs(y));
 }
 
 double L2distance( const kdtreeNode &node)const
 {
   double x = xy[0] - node.xy[0];
   double y = xy[1] - node.xy[1];
   //only supports manhattan distance
   //whatever
   return sqrt(x*x+y*y);
 }
};

//kd tree of two doubles
typedef KDTree::KDTree<DIM ,kdtreeNode> KDT2d;
bool save_eps(const char* filename, const std::vector<kdtreeNode> & sites, 
	      const double width, const double height, const double radius);
void loadPts(const char * filename, std::vector<kdtreeNode> & pts);
double dmvnorm(const int n, const gsl_vector *x, const gsl_vector *mean, const gsl_matrix *var);
int rmvnorm(const gsl_rng *r, const int n, const gsl_vector *mean, 
	    const gsl_matrix *var, gsl_vector *result);
void E(const std::vector<kdtreeNode> & pts,
  const std::vector<kdtreeNode> & sites,
  const std::vector<gsl_vector*> & mean,
  const std::vector<gsl_matrix*> & cov,
  const KDT2d & tree,
  std::vector< std::vector<double> > & expect,
  std::vector< std::vector<int> > & idx);

void M(const std::vector< std::vector<double> > & expect,
  const std::vector< std::vector<int> > & idx,
  const std::vector< kdtreeNode> & points,
  std::vector<gsl_vector*> & mean,
  std::vector<gsl_matrix*> & cov,
  std::vector<kdtreeNode> & sites
);

void normalize(std::vector<double>&v)
{
  double sum = 0;
  for(unsigned int ii = 0;ii<v.size();ii++){
    sum += v[ii];
  }
  for(unsigned int ii = 0;ii<v.size();ii++){
    v[ii]/=sum;
  }
}

template<class T>
inline T clamp(T a, T lower, T upper)
{
   if(a<lower) return lower;
   else if(a>upper) return upper;
   else return a;
}

void saveSites(const char * filename,
  std::vector<gsl_vector*> mean,
  std::vector<gsl_matrix*> cov)
{
  std::ofstream out (filename);
  out<<mean.size()<<"\n";
  for(unsigned int ii =0 ;ii<mean.size();ii++){
    double xx = gsl_vector_get(mean[ii],0);
    double yy = gsl_vector_get(mean[ii],1);
    out<<xx<<" "<<yy<<"\n";
    
    xx = gsl_matrix_get(cov[ii],0,0);
    out<<xx;
    xx = gsl_matrix_get(cov[ii],0,1);
    out<<" "<<xx<<"\n";

    xx = gsl_matrix_get(cov[ii],1,0);
    out<<xx;
    xx = gsl_matrix_get(cov[ii],1,1);
    out<<" "<<xx<<"\n";
  }
  out.close();
}

int main(int , char *[])
{
  std::vector<kdtreeNode> pts;
  std::vector<kdtreeNode> sites;
  int ITER = 20;  
  loadPts("points.txt",pts);
  loadPts("sites.txt",sites);
  KDT2d tree(pts.begin(),pts.end());
  
  int avgCapacity = pts.size()/sites.size();
  std::vector<gsl_matrix*> cov(sites.size());
  std::vector<gsl_vector*> mean(sites.size());
  std::cout<<avgCapacity<<"\n";
  //find k=capacity nearest neightbors and compute radius as covariance
  for(unsigned int ii = 0 ;ii<sites.size();ii++){
    std::pair<KDT2d::const_iterator,double> found = tree.find_nearest(sites[ii]);
    KDT2d::const_iterator it = found.first;
    int cnt = 0;
    double maxdist = 0;
    while(cnt<avgCapacity && it != tree.end()){
      double dist = sites[ii].L2distance(*it);
      if(dist>maxdist){
	maxdist = dist;
      }
      it++;
      cnt++;
    }
    double sigma2 = maxdist/5;
    sigma2*=sigma2;
    cov[ii] = gsl_matrix_alloc(DIM,DIM);
    gsl_matrix_set_zero(cov[ii]);
    gsl_matrix_set (cov[ii], 0, 0, sigma2);
    gsl_matrix_set (cov[ii], 1, 1, sigma2);
    
    mean[ii] = gsl_vector_alloc(DIM);
    gsl_vector_set_zero (mean[ii]);
    for(int jj = 0;jj<DIM;jj++){
      gsl_vector_set (mean[ii], jj, sites[ii][jj]);
    }
  }
  std::vector< std::vector<double> > expect;
  std::vector< std::vector<int> > idx;
  for(int iter = 0;iter<ITER;iter++){
    std::cout<<iter<<" iteration\n";
    E(pts, sites, mean, cov,tree,expect,idx);
    M(expect, idx, pts,mean, cov, sites);
  }
  
  saveSites("gaussian.txt",mean,cov);
  save_eps("sites.eps", sites, 1000, 1000, 5);

  std::vector<Vector3f> colors (sites.size());
  for(unsigned int ii =0 ;ii<sites.size();ii++){
    for(int jj = 0;jj<3;jj++){
      colors[ii][jj] = rand()/(float)RAND_MAX;
    }
  }
  
  return 0;
}

void E(const std::vector<kdtreeNode> & pts,
  const std::vector<kdtreeNode> & sites,
  const std::vector<gsl_vector*> & mean,
  const std::vector<gsl_matrix*> & cov,const KDT2d & tree,
  std::vector< std::vector<double> > & expect,
  std::vector< std::vector<int> > & idx)
{
  //the image is assumed to be 1000x1000
  //check points.txt to see if it is true
  double maxDist = 70;
  
  expect.resize(pts.size());
  idx.resize(pts.size());
  gsl_vector * x= gsl_vector_alloc(DIM);
  for(unsigned int ii = 0;ii<pts.size();ii++){
    expect[ii].clear();
    idx[ii].clear();
  }
  for(unsigned int ii = 0;ii<sites.size();ii++){
    std::vector<kdtreeNode> out;
    tree.find_within_range(sites[ii], maxDist, 
			   std::back_insert_iterator<std::vector<kdtreeNode> >(out));
    for(unsigned int jj = 0;jj<out.size();jj++){
      if(sites[ii].L2distance(out[jj])>maxDist){
	continue;
      }
      gsl_vector_set(x,0,out[jj][0]);
      gsl_vector_set(x,1,out[jj][1]);
      double pdf = dmvnorm(DIM, x, mean[ii],cov[ii]);
      int nodeId = out[jj].id;
      expect[nodeId].push_back(pdf);
      idx[nodeId].push_back(ii);
    }
  }

  double logPosterior = 0;
  for(unsigned int ii = 0;ii<expect.size();ii++){
    double sum = 0;
    for(unsigned int jj = 0;jj<expect[ii].size();jj++){
      sum += expect[ii][jj];
    }
    if(sum>0){
      logPosterior -= log(sum);
    }
    normalize(expect[ii]);
  }
  std::cout<<logPosterior<<"\n";
  gsl_vector_free(x);
}

void covar(const std::vector<kdtreeNode> & points,
	  const std::vector<gsl_vector*> & mean,
	  const std::vector< std::vector<double> > & expect,
	  const std::vector< std::vector<int> > & idx,
	  const std::vector<double> &wsum,
	  std::vector<gsl_matrix*> & cov,
	  int dim1, int dim2)
{
  std::vector<double> wsquare (mean.size());
  
  for(unsigned int ii = 0; ii<expect.size(); ii++){
    for(unsigned int jj =0 ;jj<expect[ii].size();jj++){
      int siteIdx = idx[ii][jj];
      double var = gsl_matrix_get(cov[siteIdx],dim1, dim2);
      double mu1 = gsl_vector_get(mean[siteIdx],dim1);
      double mu2 = gsl_vector_get(mean[siteIdx],dim2);
      var += expect[ii][jj] * (points[ii][dim1]-mu1)*(points[ii][dim2]-mu2);
      wsquare[siteIdx] += expect[ii][jj]*expect[ii][jj];
      gsl_matrix_set(cov[siteIdx],dim1, dim2, var);
    }
  }
  for(unsigned int ii = 0;ii<cov.size();ii++){
    if(wsum[ii]<=0){
      continue;
    }
    double var = gsl_matrix_get(cov[ii],dim1, dim2);
    var/=wsum[ii]-wsquare[ii]/wsum[ii];
    gsl_matrix_set(cov[ii],dim1, dim2,var);
  }
}

void M(const std::vector< std::vector<double> > & expect,
  const std::vector< std::vector<int> > & idx,
  const std::vector< kdtreeNode> & points,
  std::vector<gsl_vector*> & mean,
  std::vector<gsl_matrix*> & cov,
  std::vector<kdtreeNode> & sites
)
{
    std::vector<double> wsum;
    wsum.resize(sites.size());
    for(unsigned int ii = 0;ii< sites.size();ii++){
      gsl_vector_set_zero(mean[ii]);
    }
    
    for(unsigned int ii = 0;ii<expect.size();ii++){
      for(unsigned int jj = 0;jj<expect[ii].size();jj++){
	int siteIdx = idx[ii][jj];
	double w = expect[ii][jj];
	wsum[siteIdx] += w;
	for(int dim = 0;dim<DIM;dim++){
	  double val =  gsl_vector_get(mean[siteIdx],dim);
	  val += w*points[ii][dim];
	  gsl_vector_set(mean[siteIdx],dim,val);
	}
      }
    }

    double verifySum = 0;
    for(unsigned int ii = 0;ii<sites.size();ii++){
      verifySum+=wsum[ii];
      
      if(wsum[ii]<=0){
	continue;
      }
      for(int dim = 0;dim<DIM;dim++){
	double val =  gsl_vector_get(mean[ii],dim);
	val /= wsum[ii];
	gsl_vector_set(mean[ii],dim,val);
	
	sites[ii][dim] = val;
      }
    }
    
   // std::cout<<verifySum<<"checksum \n";
    
    for(unsigned int ii = 0;ii<cov.size();ii++){
      if(wsum[ii]>0){
	gsl_matrix_set_zero(cov[ii]);
      }
    }
    
    covar(points, mean,  expect, idx, wsum, cov,0, 0);
    covar(points, mean,  expect, idx, wsum, cov,0, 1);
    covar(points, mean,  expect, idx, wsum, cov,1, 0);
    covar(points, mean,  expect, idx, wsum, cov,1, 1);
  
}

void loadPts(const char * filename, std::vector<kdtreeNode> & pts)
{
  std::ifstream infile(filename);
  if(infile.bad()){
    std::cout<<"cannot open "<<filename<<"\n";
    return;
  }
  int nPts ;
  infile>>nPts;
  pts.resize(nPts);
  std::cout<<"loading "<<nPts<<" points\n";
  for(int ii = 0;ii<nPts;ii++){
    infile>>pts[ii][0]>>pts[ii][1];
    pts[ii].id = ii;
  }
  infile.close();
}

/********************************************************************************/
//Verified against matlab
/********************************************************************************/
double dmvnorm(const int n, const gsl_vector *x, const gsl_vector *mean, const gsl_matrix *var){
  /* multivariate normal density function    */
  /*
   *    n       dimension of the random vetor
   *    mean    vector of means of size n
   *    var     variance matrix of dimension n x n
   */
  int s;
  double ax,ay;
  gsl_vector *ym, *xm;
  gsl_matrix *work = gsl_matrix_alloc(n,n),
    *winv = gsl_matrix_alloc(n,n);
  gsl_permutation *p = gsl_permutation_alloc(n);
 
  gsl_matrix_memcpy( work, var );
  gsl_linalg_LU_decomp( work, p, &s );
  gsl_linalg_LU_invert( work, p, winv );
  ax = gsl_linalg_LU_det( work, s );
  gsl_matrix_free( work );
  gsl_permutation_free( p );
 
  xm = gsl_vector_alloc(n);
  gsl_vector_memcpy( xm, x);
  gsl_vector_sub( xm, mean );
  ym = gsl_vector_alloc(n);
  gsl_blas_dsymv(CblasUpper,1.0,winv,xm,0.0,ym);
  gsl_matrix_free( winv );
  gsl_blas_ddot( xm, ym, &ay);
  gsl_vector_free(xm);
  gsl_vector_free(ym);
  ay = exp(-0.5*ay)/sqrt( pow((2*M_PI),n)*ax );
 
  return ay;
}

/************************************************************************************************/
/************************************************************************************************/
int rmvnorm(const gsl_rng *r, const int n, const gsl_vector *mean, const gsl_matrix *var, gsl_vector *result){
  /* multivariate normal distribution random number generator */
  /*
   *    n       dimension of the random vetor
   *    mean    vector of means of size n
   *    var     variance matrix of dimension n x n
   *    result  output variable with a sigle random vector normal distribution generation
   */
  int k;
  gsl_matrix *work = gsl_matrix_alloc(n,n);
 
  gsl_matrix_memcpy(work,var);
  gsl_linalg_cholesky_decomp(work);
 
  for(k=0; k<n; k++)
    gsl_vector_set( result, k, gsl_ran_ugaussian(r) );
 
  gsl_blas_dtrmv(CblasLower, CblasNoTrans, CblasNonUnit, work, result);
  gsl_vector_add(result,mean);
 
  gsl_matrix_free(work);
 
  return 0;
}

// export sites to an EPS image
bool save_eps(const char* filename, const std::vector<kdtreeNode> & sites,
	      const double width, const double height, const double radius) {
  std::ofstream stream(filename, std::ios::out);
  if (stream.bad()) {
    return false;
  }

  stream << "%!PS-Adobe EPSF-3.0\n";
  stream << "%%HiResBoundingBox: " << 0.0 << " " << 0.0 << " " << width << " " << height << "\n";
  stream << "%%BoundingBox: " << 0 << " " << 0 << " " << static_cast<int>(width) << " " << static_cast<int>(height) << "\n";
  stream << "\n";
  stream << "%% Sites: " << sites.size() << "\n";
  stream << "\n";
  stream << "/radius { " << radius << " } def\n";
  stream << "\n";
  stream << "/p { radius 0 360 arc closepath fill } def\n";
  stream << "\n";
  stream << "0 0 0 setrgbcolor\n";
  stream << "\n";
  for (unsigned int i = 0; i < sites.size(); ++i) {
    stream << sites[i][0] << " " << sites[i][1] << " p\n";
  }
  stream << "\n";
  stream << "showpage\n";

  stream.close();
  return true;
}

